////////////////////////////////////////////////////////////////////////
// Class:       AdcThresholdRoiFinder
// Plugin Type: producer (art v3_05_01)
// File:        AdcThresholdRoiFinder_module.cc
//
// Generated at Thu Jul  2 15:42:00 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// About:  waveform ROI finder based on ADC over-threshold algorithm.
//   a. flag if a waveform is a signal waveform according to positive ADC.
//   b. some tests:
//      rebinning waveform was tested, no visible difference, this part is deleted;
//      test waveform size can be optimized according to signal width, i.e,  including some previous bins and some following bins for a peak. One may refer David Adams's similiar work on DUNE.
//   c. options for window size and stride size; a shift to last window in order to have the same window size
//   d. after filling true for roi, filling false for non-roi will narrow
//      the roi range.
//
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

#include <memory>

using namespace std;


class AdcThresholdRoiFinder;


class AdcThresholdRoiFinder : public art::EDProducer {
public:
  explicit AdcThresholdRoiFinder(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  AdcThresholdRoiFinder(AdcThresholdRoiFinder const&) = delete;
  AdcThresholdRoiFinder(AdcThresholdRoiFinder&&) = delete;
  AdcThresholdRoiFinder& operator=(AdcThresholdRoiFinder const&) = delete;
  AdcThresholdRoiFinder& operator=(AdcThresholdRoiFinder&&) = delete;

  // Required functions.
  void produce(art::Event& e) override;

private:
  // user's function
  std::vector<bool> findROI(const std::vector<float>& adcin, int view);

  // configuration from fhicl file
  std::string fRawDigitLabel;
  std::string fWireProducerLabel;
  bool fUseShort; // false: float for ADC; true: short for ADC
  std::vector<float> fThreshold;
  int fWindowSize;
  int fStrideSize;
  bool fFlagPositive;
  bool fFlagNegtive;

  int fNticks;
  int fNumStrides;
  int fLastWindowSize;
};


AdcThresholdRoiFinder::AdcThresholdRoiFinder(fhicl::ParameterSet const& p)
  : EDProducer{p}  // ,
{
  fRawDigitLabel        = p.get<std::string>("RawDigitLabel");
  fWireProducerLabel    = p.get<std::string>("WireProducerLabel");
  fUseShort             = p.get<bool>("UseShort", true);
  fThreshold            = p.get<std::vector<float>>("Threshold");
  fWindowSize           = p.get<int>("WindowSize", 200);
  fStrideSize           = p.get<int>("StrideSize", 150);
  fFlagPositive         = p.get<bool>("FlagPositive");
  fFlagNegtive          = p.get<bool>("FlagNegtive");

  if (fRawDigitLabel.empty() && fWireProducerLabel.empty()){
    throw cet::exception("AdcThresholdRoiFinder") << "Both RawDigitLabel and WireProducerLabel are empty";
  }

  if ((!fRawDigitLabel.empty()) && (!fWireProducerLabel.empty())){
    throw cet::exception("AdcThresholdRoiFinder") << "Only one of RawDigitLabel and WireProducerLabel should be set";
  }

  // DetectorPropertiesService
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService>()->DataForJob();
  fNticks = detProp.ReadOutWindowSize(); // number of clock ticks per event
  //std::cout << "Number of clock ticks per event: " << fNticks << std::endl;

  produces< std::vector<recob::Wire> >();

  if (fWindowSize > 0 && fStrideSize >0) {
    float dmn = fNticks - fWindowSize; // distance between tail edge of 1st window and last data point
    fNumStrides = std::ceil(dmn / float(fStrideSize));
    unsigned int overshoot  = fNumStrides * fStrideSize + fWindowSize - fNticks;
    fLastWindowSize = fWindowSize - overshoot;
    unsigned int numwindows = fNumStrides + 1;
    std::cout << " !!!!! AdcThresholdRoiFinder: WindowSize = " << fWindowSize
              << ", StrideSize = " << fStrideSize
              << ", dmn/StrideSize = " << dmn / fStrideSize << std::endl;
    std::cout << "   dmn = " << dmn << ", NumStrides = " << fNumStrides
              << ", overshoot = " << overshoot << ", LastWindowSize = " << fLastWindowSize
              << ", numwindows = " << numwindows << std::endl;
  }
}


void AdcThresholdRoiFinder::produce(art::Event& e)
{
  
  auto const* geo = lar::providerFrom<geo::Geometry>();

  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fRawDigitLabel, "", rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }
  
  // get Wire
  art::Handle< std::vector<recob::Wire> > wireListHandle;
  std::vector<art::Ptr<recob::Wire> > wirelist;
  if (e.getByLabel(fWireProducerLabel, wireListHandle)) {
    art::fill_ptr_vector(wirelist, wireListHandle);
  }

  // make a collection of Wires and an association set
  std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);

  // loop over wires
  // use one of rawdigitlist and  wirelist that is not empty()
  for (unsigned int ich=0; ich < (rawdigitlist.empty()?wirelist.size():rawdigitlist.size()); ++ich) {

    std::vector<float> adcvec(fNticks);
    int view = -1;

    if (!rawdigitlist.empty()) {
      const auto & digitVec = rawdigitlist[ich];
  
      view =  geo->View(digitVec->Channel());

      std::vector<short> rawadc(fNticks);
      raw::Uncompress(digitVec->ADCs(), rawadc, digitVec->GetPedestal(), digitVec->Compression());

      for (size_t j=0; j<rawadc.size(); ++j) {
        adcvec[j] = rawadc[j] - digitVec->GetPedestal(); 
      }
    }
    else if (!wirelist.empty()) {
      const auto & wire = wirelist[ich];
      const auto & signal = wire->Signal();
      
      view = wirelist[ich]->View();

      if (fUseShort) {
        for (size_t j=0; j<adcvec.size(); ++j) {
          adcvec[j] = short(signal[j]);
        }
      }
      else {
        for (size_t j=0; j<adcvec.size(); ++j) {
          adcvec[j] = signal[j];
        }
      }
    }
    
    // use ADC Over-threshold algorithm for each window
    std::vector<bool> inroi(fNticks, false);
    inroi = findROI(adcvec, view);

    recob::Wire::RegionsOfInterest_t rois(fNticks);

    std::vector<float> sigs;
    int lastsignaltick = -1;
    int roistart = -1;

    for (int j = 0; j < fNticks; ++j) {
      if (inroi[j]) {
        if (sigs.empty()) {
          sigs.push_back(adcvec[j]);
          lastsignaltick = j;
          roistart = j;
        }
        else {
          if (j != lastsignaltick + 1) {
            rois.add_range(roistart, std::move(sigs));
            sigs.clear();
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
            roistart = j;
          }
          else {
            sigs.push_back(adcvec[j]);
            lastsignaltick = j;
          }
        }
      }
    }   
    if (!sigs.empty()) { rois.add_range(roistart, std::move(sigs)); }

    if (!wirelist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, wirelist[ich]->Channel(), wirelist[ich]->View()));
    }
    else if (!rawdigitlist.empty()) {
      wirecol->emplace_back(recob::Wire(rois, rawdigitlist[ich]->Channel(), geo->View(rawdigitlist[ich]->Channel())));
    }
  
  } //loop wire ich

  e.put(std::move(wirecol));
}


std::vector<bool> AdcThresholdRoiFinder::findROI(const std::vector<float>& adcin, int view) {
  std::vector<bool> roivec(fNticks, false);
  if (int(adcin.size()) != fNticks) {return roivec;}

  // create a vector of windows
  std::vector<std::vector<float>> windowvec(fNumStrides + 1, std::vector<float>(fWindowSize, 0.));

  // fill each window with adc values
  unsigned int j1, j2, k;
  for (int i = 0; i < fNumStrides; i++) {
    j1 = i * fStrideSize;
    j2 = j1 + fWindowSize;
    k = 0;
    for (unsigned int j = j1; j < j2; j++) {
      windowvec[i][k] = adcin[j];
      k++;
    }
  }
  // last window as a special case: still has a size of fWindowSize, by shifting backward.
  j2 = fNticks;
  j1 = j2 - fWindowSize;
  k = 0;
  for (unsigned int j = j1; j < j2; j++) {
    windowvec[fNumStrides][k] = adcin[j];
    k++;
  }

  // set to true for all bins in a window if it is identified as signal
  for (int i = 0; i < fNumStrides; i++) {
    bool isroi = false;
    for (int k = 0; k < fWindowSize; ++k) {
      if ( (fFlagPositive && windowvec[i][k] > fThreshold[view]) || (fFlagNegtive && windowvec[i][k] < - fThreshold[view]) ) {
        isroi = true;
        break;
      }    
    }
    if (isroi) {
      std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, true);
    }
    // filling false here can narrow the roi range
    else {
      std::fill_n(roivec.begin() + i * fStrideSize, fWindowSize, false);
    }
  }
  // last window as a special case
  bool isroi_lastwindow = false;
  for (int k = 0; k < fWindowSize; ++k) {
    if ( (fFlagPositive && windowvec[fNumStrides][k] > fThreshold[view]) || (fFlagNegtive && windowvec[fNumStrides][k] < - fThreshold[view]) ) {
      isroi_lastwindow = true;
      break;
    }
  }
  if (isroi_lastwindow) {
    std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, true);
  }
  // filling false here can narrow the roi range
  else {
    std::fill_n(roivec.begin() + fNticks - fWindowSize, fWindowSize, false);
  }

  return roivec;

} // findROI


DEFINE_ART_MODULE(AdcThresholdRoiFinder)
